package org.apache.poi.xssf.usermodel.bugs;

import org.junit.Assert;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.xssf.usermodel.XSSFFormulaEvaluator;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.junit.Test;

/**
 * Usually I need to combine the Excel functions: INDIRECT, ADDRESS, ROW and COLUMN
 * in order to avoid reference problems when deleting rows in columns with formulas like:
 *  
 *     IF(INDIRECT(ADDRESS(ROW()-1, COLUMN())) > 0, INDIRECT(ADDRESS(ROW()-1, COLUMN())), 0)
 * 
 * Here we check if the value of the cell before the current row is valid (in this case greater than zero) 
 * and if so we return that value. This works ok on Excel but when evaluating in POI I get an exception 
 * with message: "row index may not be negative", this only happens with rows when ROW()-1 not when COLUMN()-1.
 * Here we find two test cases showing the case with ROW()-1 and the case with COLUMN()-1 
 * 
 * @author Raul Contreras
 *
 */
public class Bug56820Test {
 
	/**
	 * Sums 2 plus the cell at the left, indirectly to avoid reference 
	 * problems when deleting columns, conditionally to stop recursion 
	 */
	private static final String FORMULA1 = 
			"IF( INDIRECT( ADDRESS( ROW(), COLUMN()-1 ) ) = 0, 0,"
					+ "INDIRECT( ADDRESS( ROW(), COLUMN()-1 ) ) ) + 2";
	
	/**
	 * Sums 2 plus the upper cell, indirectly to avoid reference 
	 * problems when deleting rows, conditionally to stop recursion 
	 */
	private static final String FORMULA2 = 
			"IF( INDIRECT( ADDRESS( ROW()-1, COLUMN() ) ) = 0, 0,"
					+ "INDIRECT( ADDRESS( ROW()-1, COLUMN() ) ) ) + 2";
	
	
	/**
	 * Expected:
	 * 
	 * [  0][  2][  4]
	 */
	@Test public void testBug56820_Formula1() {
		
		XSSFWorkbook wb = new XSSFWorkbook();
		XSSFFormulaEvaluator evaluator = new XSSFFormulaEvaluator(wb);
		Sheet sh = wb.createSheet();
		
		sh.createRow(0).createCell(0).setCellValue(0.0d);
		Cell formulaCell1 = sh.getRow(0).createCell(1);
		Cell formulaCell2 = sh.getRow(0).createCell(2);
		formulaCell1.setCellFormula(FORMULA1);
		formulaCell2.setCellFormula(FORMULA1);
		
		double A1 = evaluator.evaluate(formulaCell1).getNumberValue();
		double A2 = evaluator.evaluate(formulaCell2).getNumberValue(); 
		
		Assert.assertEquals(2, A1, 0);
		Assert.assertEquals(4, A2, 0);  //<-- FAILS EXPECTATIONS
	}
	
	
	
	/**
	 * Expected:
	 * 
	 * [  0] <- number
	 * [  2] <- formula
	 * [  4] <- formula
	 */
	@Test public void testBug56820_Formula2() {
		
		XSSFWorkbook wb = new XSSFWorkbook();
		XSSFFormulaEvaluator evaluator = new XSSFFormulaEvaluator(wb);
		Sheet sh = wb.createSheet();
		
		sh.createRow(0).createCell(0).setCellValue(0.0d);
		Cell formulaCell1 = sh.createRow(1).createCell(0);
		Cell formulaCell2 = sh.createRow(2).createCell(0);
		formulaCell1.setCellFormula(FORMULA2);
		formulaCell2.setCellFormula(FORMULA2);
		
		double A1 = evaluator.evaluate(formulaCell1).getNumberValue();
		double A2 = evaluator.evaluate(formulaCell2).getNumberValue(); //<-- FAILS EVALUATION
		
		Assert.assertEquals(2, A1, 0);
		Assert.assertEquals(4, A2, 0);
	}
}
